@@ -18,7 +18,7 @@ where the curve is traced, the axes are labeled and the scan rate is provided.
 This SVG file can then be analyzed by this class to produce the coordinates
 corresponding to the original measured values.
 
-A more detailed description on preparing the SVG files is provieded in the :class:`CV`
+A more detailed description on preparing the SVG files is provided in the :class:`CV`
 or ...
 
 TODO:: Link to workflow.md (see issue #73)
@@ -29,9 +29,9 @@ For the documentation below, the path of a CV is presented simply as a line.
 # ********************************************************************
 #  This file is part of svgdigitizer.
 #
-#        Copyright (C) 2021-2022 Albert Engstfeld
+#        Copyright (C) 2021-2023 Albert Engstfeld
 #        Copyright (C) 2021      Johannes Hermann
-#        Copyright (C) 2021-2022 Julian Rüth
+#        Copyright (C) 2021-2023 Julian Rüth
 #        Copyright (C) 2021      Nicolas Hörmann
 #
 #  svgdigitizer is free software: you can redistribute it and/or modify
@@ -48,11 +48,13 @@ For the documentation below, the path of a CV is presented simply as a line.
 #  along with svgdigitizer. If not, see <https://www.gnu.org/licenses/>.
 # ********************************************************************
 import logging
-from functools import cache
+from functools import cached_property
 
 import matplotlib.pyplot as plt
 from astropy import units as u
 
+from svgdigitizer.exceptions import SVGAnnotationError
+
 logger = logging.getLogger("cv")
 
 
@@ -74,7 +76,7 @@ class CV:
     It requires:
 
     * | that the x-axis is labeled with U or E (V) and the y-axis
-      | is labeld by I (A) or j (A / cm2)
+      | is labeled by I (A) or j (A / cm2)
     * | that the label of the second point (furthest from the origin)
       | on the x- or y-axis contains a value and a unit
       | such as ``<text>j2: 1 mA / cm2</text>`` or ``<text>E2: 1 mV</text>``.
@@ -87,7 +89,7 @@ class CV:
 
     * | A comment describing the data, i.e.,
       | ``<text>comment: noisy data</text>``
-    * | Other measurements linked to this measurement or performed simultanouesly, i.e.,
+    * | Other measurements linked to this measurement or performed simultaneously, i.e.,
       | ``<text>linked: SXRD, DEMS</text>``
     * | A list of tags describing the content of a plot, i.e.,
       | ``<text>tags: BCV, HER, OER</text>``
@@ -104,7 +106,7 @@ class CV:
         ... <svg>
         ...   <g>
         ...     <path d="M 0 100 L 100 0" />
-        ...     <text x="0" y="0">curve: 0</text>
+        ...     <text x="0" y="0">curve: solid</text>
         ...   </g>
         ...   <g>
         ...     <path d="M 0 200 L 0 100" />
@@ -147,23 +149,24 @@ class CV:
 
     The properties of the original plot and the dataframe can be returned as a dict::
 
-        >>> cv.metadata  # doctest: +NORMALIZE_WHITESPACE
-        {'experimental': {'tags': ['BCV', 'HER', 'OER']},
-         'source': {'figure': '2b', 'curve': '0'},
-         'figure description': {'version': 1,
-          'type': 'digitized',
-          'simultaneous measurements': ['SXRD', 'SHG'],
-          'measurement type': 'CV',
-          'scan rate': {'value': 50.0, 'unit': 'V / s'},
-          'fields': [{'name': 'E', 'unit': 'mV', 'orientation': 'x',
-                    'reference': 'RHE'},
-                    {'name': 'j', 'unit': 'uA / cm2', 'orientation': 'y'}],
-                    'comment': 'noisy data'},
-          'data description': {'version': 1, 'type': 'digitized',
-                              'measurement type': 'CV', 'fields':
-                              [{'name': 'E', 'unit': 'V', 'reference': 'RHE'},
-                              {'name': 'j', 'unit': 'A / m2'},
-                              {'name': 't', 'unit': 's'}]}}
+        >>> cv.metadata  == \
+        ... {'experimental': {'tags': ['BCV', 'HER', 'OER']},
+        ...  'source': {'figure': '2b', 'curve': 'solid'},
+        ...  'figure description': {'version': 1,
+        ...                         'type': 'digitized',
+        ...                         'simultaneous measurements': ['SXRD', 'SHG'],
+        ...                         'measurement type': 'CV',
+        ...                         'scan rate': {'value': 50.0, 'unit': 'V / s'},
+        ...                         'fields': [{'name': 'E','unit': 'mV', 'orientation': 'x', 'reference': 'RHE', 'type': 'number'},
+        ...                                    {'name': 'j', 'unit': 'uA / cm2', 'orientation': 'y', 'type': 'number'}],
+        ...                         'comment': 'noisy data'},
+        ...  'data description': {'version': 1,
+        ...                       'type': 'digitized',
+        ...                       'measurement type': 'CV',
+        ...                       'fields': [{'name': 'E', 'type': 'number', 'unit': 'V', 'reference': 'RHE'},
+        ...                                  {'name': 'j', 'type': 'number', 'unit': 'A / m2'},
+        ...                                  {'name': 't', 'type': 'number', 'unit': 's'}]}}
+        True
 
     """
 
@@ -185,6 +188,10 @@ class CV:
             >>> svg = SVG(StringIO(r'''
             ... <svg>
             ...   <g>
+            ...     <path d="M 0 100 L 100 0" />
+            ...     <text x="0" y="0">curve: 0</text>
+            ...   </g>
+            ...   <g>
             ...     <path d="M 0 200 L 0 100" />
             ...     <text x="0" y="200">E1: 0 V vs. RHE</text>
             ...   </g>
@@ -216,6 +223,10 @@ class CV:
             >>> svg = SVG(StringIO(r'''
             ... <svg>
             ...   <g>
+            ...     <path d="M 0 100 L 100 0" />
+            ...     <text x="0" y="0">curve: 0</text>
+            ...   </g>
+            ...   <g>
             ...     <path d="M 0 200 L 0 100" />
             ...     <text x="0" y="200">j1: 0 uA / cm2</text>
             ...   </g>
@@ -237,17 +248,22 @@ class CV:
             >>> cv.voltage_dimension
             Traceback (most recent call last):
             ...
-            Exception: The voltage must be on the x-axis.
+            svgdigitizer.exceptions.SVGAnnotationError: The voltage must be on the x-axis in the SVG.
 
         """
         dimensions = list(set(["E", "U"]).intersection(self.svgplot.schema.field_names))
 
         if len(dimensions) == 1:
-            if self.svgplot.schema.get_field(dimensions[0])["orientation"] == "x":
+            if (
+                self.svgplot.schema.get_field(dimensions[0]).custom["orientation"]
+                == "x"
+            ):
                 return dimensions[0]
-            raise Exception("The voltage must be on the x-axis.")
+            raise SVGAnnotationError("The voltage must be on the x-axis in the SVG.")
 
-        raise Exception("No voltage axis or more than one voltage axis found.")
+        raise SVGAnnotationError(
+            "No voltage axis or more than one voltage axis found in the SVG."
+        )
 
     @property
     def current_dimension(self):
@@ -264,6 +280,10 @@ class CV:
             >>> svg = SVG(StringIO(r'''
             ... <svg>
             ...   <g>
+            ...     <path d="M 0 100 L 100 0" />
+            ...     <text x="0" y="0">curve: 0</text>
+            ...   </g>
+            ...   <g>
             ...     <path d="M 0 200 L 0 100" />
             ...     <text x="0" y="200">E1: 0 V vs. RHE</text>
             ...   </g>
@@ -289,11 +309,16 @@ class CV:
         dimensions = list(set(["I", "j"]).intersection(self.svgplot.schema.field_names))
 
         if len(dimensions) == 1:
-            if self.svgplot.schema.get_field(dimensions[0])["orientation"] == "y":
+            if (
+                self.svgplot.schema.get_field(dimensions[0]).custom["orientation"]
+                == "y"
+            ):
                 return dimensions[0]
-            raise Exception("The current must be on the x-axis.")
+            raise SVGAnnotationError("The current must be on the y-axis in the SVG.")
 
-        raise Exception("No current axis or more than one current axis found.")
+        raise SVGAnnotationError(
+            "No current axis or more than one current axis found in the SVG."
+        )
 
     @property
     def data_schema(self):
@@ -313,6 +338,10 @@ class CV:
             >>> svg = SVG(StringIO(r'''
             ... <svg>
             ...   <g>
+            ...     <path d="M 0 100 L 100 0" />
+            ...     <text x="0" y="0">curve: 0</text>
+            ...   </g>
+            ...   <g>
             ...     <path d="M 0 200 L 0 100" />
             ...     <text x="0" y="200">E1: 0 V vs. RHE</text>
             ...   </g>
@@ -332,9 +361,9 @@ class CV:
             ... </svg>'''))
             >>> cv = CV(SVGPlot(svg))
             >>> cv.data_schema  # doctest: +NORMALIZE_WHITESPACE
-            {'fields': [{'name': 'E', 'unit': 'V', 'reference': 'RHE'},
-                        {'name': 'j', 'unit': 'A / m2'},
-                        {'name': 't', 'unit': 's'}]}
+            {'fields': [{'name': 'E', 'type': 'number', 'unit': 'V', 'reference': 'RHE'},
+                        {'name': 'j', 'type': 'number', 'unit': 'A / m2'},
+                        {'name': 't', 'type': 'number', 'unit': 's'}]}
 
         An SVG with a current axis with dimension I and
         a voltage axis with dimension U.::
@@ -346,6 +375,10 @@ class CV:
             >>> svg = SVG(StringIO(r'''
             ... <svg>
             ...   <g>
+            ...     <path d="M 0 100 L 100 0" />
+            ...     <text x="0" y="0">curve: 0</text>
+            ...   </g>
+            ...   <g>
             ...     <path d="M 0 200 L 0 100" />
             ...     <text x="0" y="200">U1: 0 V</text>
             ...   </g>
@@ -365,28 +398,30 @@ class CV:
             ... </svg>'''))
             >>> cv = CV(SVGPlot(svg))
             >>> cv.data_schema  # doctest: +NORMALIZE_WHITESPACE
-            {'fields': [{'name': 'U', 'unit': 'V', 'reference': 'unknown'},
-                        {'name': 'I', 'unit': 'A'},
-                        {'name': 't', 'unit': 's'}]}
+            {'fields': [{'name': 'U', 'type': 'number', 'unit': 'V', 'reference': 'unknown'},
+                        {'name': 'I', 'type': 'number', 'unit': 'A'},
+                        {'name': 't', 'type': 'number', 'unit': 's'}]}
 
         """
+        from frictionless import fields
 
         schema = self.figure_schema
 
-        schema.get_field(self.voltage_dimension)["unit"] = "V"
-        del schema.get_field(self.voltage_dimension)["orientation"]
+        schema.get_field(self.voltage_dimension).custom["unit"] = "V"
+        del schema.get_field(self.voltage_dimension).custom["orientation"]
         if self.current_dimension == "I":
-            schema.get_field(self.current_dimension)["unit"] = "A"
+            schema.get_field(self.current_dimension).custom["unit"] = "A"
         elif self.current_dimension == "j":
-            schema.get_field(self.current_dimension)["unit"] = "A / m2"
+            schema.get_field(self.current_dimension).custom["unit"] = "A / m2"
         else:
-            raise Exception(
-                "None of the axis labels has a dimension current 'I' or current density 'j'."
+            raise SVGAnnotationError(
+                "None of the axis labels in the SVG have a dimension current 'I' or current density 'j'."
             )
 
-        del schema.get_field(self.current_dimension)["orientation"]
-        schema.add_field(name="t")
-        schema.get_field("t")["unit"] = "s"
+        del schema.get_field(self.current_dimension).custom["orientation"]
+
+        schema.add_field(fields.NumberField(name="t"))
+        schema.update_field("t", {"unit": "s"})
 
         return schema
 
@@ -395,7 +430,7 @@ class CV:
         # TODO: use intersphinx to link Schema and Fields to frictionless docu (see #151).
         r"""
         A frictionless `Schema` object, including a `Fields` object
-        describing the voltage and current axis of the originlal plot
+        describing the voltage and current axis of the original plot
         including original units. The reference electrode of the
         potential/voltage axis is also given (if available).
 
@@ -408,6 +443,10 @@ class CV:
             >>> svg = SVG(StringIO(r'''
             ... <svg>
             ...   <g>
+            ...     <path d="M 0 100 L 100 0" />
+            ...     <text x="0" y="0">curve: 0</text>
+            ...   </g>
+            ...   <g>
             ...     <path d="M 0 200 L 0 100" />
             ...     <text x="0" y="200">E1: 0 V vs. RHE</text>
             ...   </g>
@@ -427,8 +466,8 @@ class CV:
             ... </svg>'''))
             >>> cv = CV(SVGPlot(svg))
             >>> cv.figure_schema  # doctest: +NORMALIZE_WHITESPACE
-            {'fields': [{'name': 'E', 'unit': 'V', 'orientation': 'x', 'reference': 'RHE'},
-                        {'name': 'j', 'unit': 'uA / cm2', 'orientation': 'y'}]}
+            {'fields': [{'name': 'E', 'type': 'number', 'unit': 'V', 'orientation': 'x', 'reference': 'RHE'},
+                        {'name': 'j', 'type': 'number', 'unit': 'uA / cm2', 'orientation': 'y'}]}
 
         """
         import re
@@ -437,19 +476,23 @@ class CV:
 
         pattern = r"^(?P<unit>.+?)? *(?:(?:@|vs\.?) *(?P<reference>.+))?$"
         match = re.match(
-            pattern, schema.get_field(self.voltage_dimension)["unit"], re.IGNORECASE
+            pattern,
+            schema.get_field(self.voltage_dimension).custom["unit"],
+            re.IGNORECASE,
         )
 
-        schema.get_field(self.voltage_dimension)["unit"] = match[1]
-        schema.get_field(self.voltage_dimension)["reference"] = match[2] or "unknown"
+        schema.update_field(
+            self.voltage_dimension,
+            {"unit": match[1], "reference": match[2] or "unknown"},
+        )
 
         return schema
 
-    @property
-    @cache
+    @cached_property
     def figure_label(self):
         r"""
-        An identifier of the plot to distinguish it from other figures on the same page.
+        An identifier of the plot to distinguish it from other
+        figures on the same page.
 
         The figure name is read from a ``<text>`` in the SVG file
         such as ``<text>figure: 2b</text>``.
@@ -502,13 +545,14 @@ class CV:
 
         return figure_labels[0].label
 
-    @property
-    @cache
+    @cached_property
     def curve_label(self):
         r"""
-        A descriptive label for this curve to distinguish it from other curves in the same plot.
+        A descriptive label for this curve to distinguish it
+        from other curves in the same plot.
 
-        The curve label read from a ``<text>`` in the SVG file such as ``<text>curve: solid line</text>``.
+        The curve label read from a ``<text>`` in the SVG file
+        such as ``<text>curve: solid line</text>``.
 
         EXAMPLES::
 
@@ -556,8 +600,7 @@ class CV:
 
         return curve_labels[0].label
 
-    @property
-    @cache
+    @cached_property
     def scan_rate(self):
         r"""
         Return the scan rate of the plot.
@@ -601,22 +644,21 @@ class CV:
         )
 
         if len(rates) > 1:
-            raise ValueError(
-                "Multiple text fields with a scan rate were provided in the SVG file. Remove all but one."
+            raise SVGAnnotationError(
+                "Multiple text fields with a scan rate were provided in the SVG. Remove all but one."
             )
 
         if not rates:
             rate = self._metadata.get("figure description", {}).get("scan rate", {})
 
             if "value" not in rate or "unit" not in rate:
-                raise ValueError("No text with scan rate found in the SVG.")
+                raise SVGAnnotationError("No text with scan rate found in the SVG.")
 
             return float(rate["value"]) * u.Unit(str(rate["unit"]))
 
         return float(rates[0].value) * u.Unit(rates[0].unit)
 
-    @property
-    @cache
+    @cached_property
     def df(self):
         # TODO: Add a more meaningful curve that reflects the shape of a cyclic voltammogram and which is displayed in the documentation (see issue #73).
         r"""
@@ -733,7 +775,7 @@ class CV:
 
         """
         voltage = 1 * u.Unit(
-            self.figure_schema.get_field(self.voltage_dimension)["unit"]
+            self.figure_schema.get_field(self.voltage_dimension).custom["unit"]
         )
         # Convert the axis unit to SI unit V and use the value
         # to convert the potential values in the df to V
@@ -778,7 +820,7 @@ class CV:
 
         """
         current = 1 * u.Unit(
-            self.figure_schema.get_field(self.current_dimension)["unit"]
+            self.figure_schema.get_field(self.current_dimension).custom["unit"]
         )
 
         # Distinguish whether the y data is current ('A') or current density ('A / cm2')
@@ -847,7 +889,7 @@ class CV:
             ... <svg>
             ...   <g>
             ...     <path d="M 0 100 L 100 0" />
-            ...     <text x="0" y="0">curve: 0</text>
+            ...     <text x="0" y="0">curve: solid</text>
             ...   </g>
             ...   <g>
             ...     <path d="M 0 200 L 0 100" />
@@ -879,20 +921,19 @@ class CV:
         plt.xlabel(
             self.voltage_dimension
             + " ["
-            + str(self.data_schema.get_field(self.voltage_dimension)["unit"])
+            + str(self.data_schema.get_field(self.voltage_dimension).custom["unit"])
             + " vs. "
-            + self.data_schema.get_field(self.voltage_dimension)["reference"]
+            + self.data_schema.get_field(self.voltage_dimension).custom["reference"]
             + "]"
         )
         plt.ylabel(
             self.current_dimension
             + " ["
-            + str(self.data_schema.get_field(self.current_dimension)["unit"])
+            + str(self.data_schema.get_field(self.current_dimension).custom["unit"])
             + "]"
         )
 
-    @property
-    @cache
+    @cached_property
     def comment(self):
         r"""
         Return a comment describing the plot.
@@ -1016,7 +1057,7 @@ class CV:
 
         """
         linked = self.svgplot.svg.get_texts(
-            "(?:simultaneous measuerment|linked|linked measurement): (?P<value>.*)"
+            "(?:simultaneous measurement|linked|linked measurement): (?P<value>.*)"
         )
 
         if len(linked) > 1:
@@ -1125,23 +1166,22 @@ class CV:
             ...   <text x="-200" y="730">tags: BCV, HER, OER</text>
             ... </svg>'''))
             >>> cv = CV(SVGPlot(svg))
-            >>> cv.metadata  # doctest: +NORMALIZE_WHITESPACE
-            {'experimental': {'tags': ['BCV', 'HER', 'OER']},
-             'source': {'figure': '2b', 'curve': '0'},
-             'figure description': {'version': 1,
-             'type': 'digitized',
-             'simultaneous measurements': ['SXRD', 'SHG'],
-             'measurement type': 'CV',
-             'scan rate': {'value': 50.0, 'unit': 'V / s'},
-             'fields': [{'name': 'E', 'unit': 'mV', 'orientation': 'x',
-                        'reference': 'RHE'},
-                        {'name': 'j', 'unit': 'uA / cm2', 'orientation': 'y'}],
-                        'comment': 'noisy data'},
-             'data description': {'version': 1, 'type': 'digitized',
-                                  'measurement type': 'CV', 'fields':
-                                  [{'name': 'E', 'unit': 'V', 'reference': 'RHE'},
-                                  {'name': 'j', 'unit': 'A / m2'},
-                                  {'name': 't', 'unit': 's'}]}}
+            >>> cv.metadata == \
+            ... {'experimental': {'tags': ['BCV', 'HER', 'OER']},
+            ...  'source': {'figure': '2b', 'curve': '0'},
+            ...  'figure description': {'version': 1,
+            ...  'type': 'digitized',
+            ...  'simultaneous measurements': ['SXRD', 'SHG'],
+            ...  'measurement type': 'CV',
+            ...  'scan rate': {'value': 50.0, 'unit': 'V / s'},
+            ...  'fields': [{'name': 'E', 'type': 'number', 'orientation': 'x', 'reference': 'RHE', 'unit': 'mV'},
+            ...             {'name': 'j', 'type': 'number', 'orientation': 'y', 'unit': 'uA / cm2'}],
+            ...  'comment': 'noisy data'},
+            ...  'data description': {'version': 1, 'type': 'digitized', 'measurement type': 'CV', 'fields':
+            ...                       [{'name': 'E', 'type': 'number', 'reference': 'RHE', 'unit': 'V'},
+            ...                       {'name': 'j', 'type': 'number', 'unit': 'A / m2'},
+            ...                       {'name': 't', 'type': 'number', 'unit': 's'}]}}
+            True
 
         """
         metadata = {
@@ -1161,17 +1201,28 @@ class CV:
                     "value": float(self.scan_rate.value),
                     "unit": str(self.scan_rate.unit),
                 },
-                "fields": self.figure_schema.fields,
+                "fields": self.figure_schema.to_dict()["fields"],
                 "comment": self.comment,
             },
             "data description": {
                 "version": 1,
                 "type": "digitized",
                 "measurement type": "CV",
-                "fields": self.data_schema.fields,
+                "fields": self.data_schema.to_dict()["fields"],
             },
         }
 
         from mergedeep import merge
 
         return merge({}, self._metadata, metadata)
+
+
+# Ensure that cached properties are tested, see
+# https://stackoverflow.com/questions/69178071/cached-property-doctest-is-not-detected/72500890#72500890
+__test__ = {
+    "CV.figure_label": CV.figure_label,
+    "CV.curve_label": CV.curve_label,
+    "CV.scan_rate": CV.scan_rate,
+    "CV.df": CV.df,
+    "CV.comment": CV.comment,
+}
